/* This file is part of ssp                             https://lila.oss/ssp
 * Copyright (C) 2023 Olivier Brunel                          jjk@jjacky.com */
/* SPDX-License-Identifier: GPL-2.0-only */
#include <errno.h>
#include <limb/base32.h>
#include <limb/bytestr.h>
#include <limb/command.h>
#include <limb/esc.h>
#include <limb/exitcode.h>
#include <limb/hasher.h> /* ALGO_* */
#include <limb/loadopt.h>
#include <limb/output.h>
#include <limb/stralloc.h>
#include <limb/u16.h>
#include "ssp.h"

struct add {
    stralloc sa;
    const char *secret;
    size_t cmtoff;
    size_t entoff;
    struct otp otp;
};

int
validate_algo(int *first, const char *data, size_t dlen)
{
    dbg("validating algo ", ESC, PMLEN(data, dlen), ESC);
    int algo = byte_get_match(first, data, dlen, (const char **) algos);
    if (algo < 0)
        warn("invalid algorithm: ", PMLEN(data, dlen));
    return algo;
}

int
validate_digits(const char *data, size_t dlen)
{
    dbg("validating digits ", ESC, PMLEN(data, dlen), ESC);
    int digits = *data - '0';
    if (dlen > 1 || digits < 5 || digits > 9) {
        warn("invalid number of digits, must be between 5 and 9: ", PMLEN(data, dlen));
        return -1;
    }
    return digits;
}

int
validate_counter(u64 *c, const char *data, size_t dlen)
{
    dbg("validating counter ", ESC, PMLEN(data, dlen), ESC);
    if (u64_scan(c, data) != dlen) {
        warn("invalid counter value: ", PMLEN(data, dlen));
        return -1;
    }
    return 0;
}

int
validate_precision(const char *data, size_t dlen)
{
    dbg("validating precision ", ESC, PMLEN(data, dlen), ESC);
    u16 u;
    if (u16_scan(&u, data) != dlen || u < 10 || u > 59) {
        warn("invalid precision argument, must be between 10 and 59: ", PMLEN(data, dlen));
        return -1;
    }
    return u;
}

enum {
    ARGID_ENTRY = OPTID_FIRST,
    ARGID_SECRET
};

COMMAND(add, "Add a new entry", "[OPTION..] <entry> <secret>",
" -a, --algo ALGO                       Set ALGO as hashing algorithm [sha1]\n"
"                                       Use 'list' to list available algorithms\n"
" -C, --comments COMMENTS               Set COMMENTS as entry's comments\n"
" -c, --counter NUM                     Set to counter-based (HOTP) with counter of NUM [1]\n"
" -d, --digits NUM                      Set to return an OTP of NUM digits [6]\n"
" -t, --time[=SECS]                     Set to time-based (TOTP) with precision of SECS [30]\n"
);

static int
parse_cmdline(int argc, const char *argv[], const char usage[], struct add *ctx)
{
    const struct option options[] = {
        OPTION_ARG_REQ ('a', "algo",                    0,          OPTID_SHORTOPT),
        OPTION_ARG_REQ ('C', "comments",                OPT_PATH,   OPTID_SHORTOPT),
        OPTION_ARG_REQ ('c', "counter",                 0,          OPTID_SHORTOPT),
        OPTION_ARG_REQ ('d', "digits",                  0,          OPTID_SHORTOPT),
        OPTION_ARG_OPT ('t', "time",                    0,          OPTID_SHORTOPT),
        LOADOPT_ARGUMENTS,
        ARGUMENT_REQ( "entry",                          OPT_PATH,   ARGID_ENTRY),
        ARGUMENT_REQ( "secret",                         0,          ARGID_SECRET),
        LOADOPT_DONE
    };
    struct loadopt lo = LOADOPT_ZERO;

    int c, r;
    while ((c = loadopt(&ctx->sa, argc, argv, options, 0, NULL, 0, &lo))) switch (c) {
        case 'a':
            {
                if (!strcmp(LO_ARG(&lo), "list")) {
                    out("Available algorithms:");
                    for (r = 0; algos[r]; ++r)
                        out("- ", algos[r]);
                    diecmdusage(0, usage, &command_add);
                }

                int first = -1;
                r = validate_algo(&first, LO_ARG(&lo), strlen(LO_ARG(&lo)));
                if (r < 0) {
                    if (first > -1)
                        list_matches(obuffer_1, OLVL_NORMAL, "did you mean ",
                                     NULL, " or ", " ?", LO_ARG(&lo), strlen(LO_ARG(&lo)),
                                     first, (const char **) algos);
                    diecmdusage(EX_USAGE, usage, &command_add);
                }
                ctx->otp.algo = r;
            }
            break;
        case 'C':
            ctx->cmtoff = LO_OFF(&lo);
            break;
        case 'c':
            if (validate_counter(&ctx->otp.n, LO_ARG(&lo), strlen(LO_ARG(&lo))) < 0)
                diecmdusage(EX_USAGE, usage, &command_add);
            ctx->otp.type = TYPE_COUNTER;
            break;
        case 'd':
            r = validate_digits(LO_ARG(&lo), strlen(LO_ARG(&lo)));
            if (r < 0)
                diecmdusage(EX_USAGE, usage, &command_add);
            ctx->otp.digits = r;
            break;
        case 't':
            if (!LO_ARG(&lo)) {
                ctx->otp.n = 30;
            } else {
                r = validate_precision(LO_ARG(&lo), strlen(LO_ARG(&lo)));
                if (r < 0)
                    diecmdusage(EX_USAGE, usage, &command_add);
                ctx->otp.n = r;
            }
            ctx->otp.type = TYPE_TIME;
            break;

        case ARGID_ENTRY:
            ctx->entoff = LO_OFF(&lo);
            break;
        case ARGID_SECRET:
            ctx->secret = LO_ARG(&lo);
            break;

        case -1:
            diecmdusage(EX_USAGE, usage, &command_add);
        default:
            die(EX_SOFTWARE, "unexpected return value ", PMINT(c), " from loadopt");
    };

    return LO_CUR(&lo);
}

int
add_main(int argc, const char *argv[], const char *env[], const char usage[], void *ssp_)
{
    struct ssp *ssp = ssp_;
    struct add ctx = { .sa = STRALLOC_ZERO, .otp = OTP_DEFAULT,
        .secret = NULL, .cmtoff = (size_t) -1, .entoff = 0 };

    (void) env;
    parse_cmdline(argc, argv, usage, &ctx);

    ssize_t l = base32_scan(NULL, ctx.secret, strlen(ctx.secret), 0);
    if (l < 0)
        dief(EX_DATA_ERR, "invalid secret: Not in base32");
    if (!l)
        dief(EX_DATA_ERR, "secret is empty");

    size_t clen = (ctx.cmtoff != (size_t) -1) ? strlen(ctx.sa.s + ctx.cmtoff) : 0;
    size_t otplen = sizeof(ctx.otp) + l + clen + 1;

    if (!stralloc_readyplus(&ctx.sa, otplen))
        diefusys(EX_TEMPFAIL, "add entry ", ESC, ctx.sa.s + ctx.entoff, ESC);

    struct otp *p = (struct otp *) (ctx.sa.s + ctx.sa.len);
    ctx.sa.len += otplen;
    memcpy(p, &ctx.otp, sizeof(ctx.otp));
    p->slen = l;
    base32_scan(p->data, ctx.secret, strlen(ctx.secret), 0);
    if (clen)
        memcpy(p->data + l, ctx.sa.s + ctx.cmtoff, clen + 1);
    else
        p->data[l] = 0;

    if (open_db(ssp, 0) < 0)
        diefu(exitcode_from_errno(errno), "open database");

    const char *entry = ctx.sa.s + ctx.entoff;
    quiet("Adding entry ", ESC, entry, ESC, "...");
    rebuild_db(NULL, entry, p, ssp);

    dbg("showing added entry");
    close_db(ssp);
    const char *av[] = { NULL, entry };
    run_command("show", sizeof(av) / sizeof(*av), av, env, ssp);

    stralloc_free(&ctx.sa);
    return 0;
}
