/* This file is part of ssp                             https://lila.oss/ssp
 * Copyright (C) 2023 Olivier Brunel                          jjk@jjacky.com */
/* SPDX-License-Identifier: GPL-2.0-only */
#include <errno.h>
#include <fcntl.h> /* AT_FDCWD */
#include <string.h>
#include <unistd.h>
#include <limb/bytestr.h>
#include <limb/cdb.h>
#include <limb/cdbmake.h>
#include <limb/endian.h>
#include <limb/exitcode.h>
#include <limb/hasher.h>
#include <limb/output.h>
#include <limb/shldata-rw.h>
#include <limb/stralloc.h>
#include <limb/term.h>
#include <limb/u32.h>
#include "ssp.h"

static ssize_t
get_pwd(const char *prompt, struct ssp *ctx)
{
    ssize_t plen;
    if (!*ctx->pwd) {
        dbg("asking for password");
        if (prompt) adde("[", PUTMSG_TOGGLE_ESC, db_file(ctx), PUTMSG_TOGGLE_ESC, "] ");
        plen = ask_password(ctx->pwd, sizeof(ctx->pwd), prompt);
    } else {
        dbg("password known");
        plen = strlen(ctx->pwd);
    }
    return plen;
}

int
open_db(struct ssp *ctx, int is_needed)
{
    dbg("opening database");

    if (ctx->cdb.map) {
        warnu("open ", ESC, db_file(ctx), ESC, ": database already opened");
        errno = EALREADY;
        return -1;
    }

    if (access(db_file(ctx), R_OK) < 0) {
        if (errno != ENOENT || is_needed) {
            warnusys("open ", ESC, db_file(ctx), ESC);
            return -1;
        }
        ctx->cdb = cdb_zero;
        ctx->cdboff = ctx->sa.len;
        return 0;
    }

    size_t off  = ctx->sa.len;
    ssize_t plen;

    plen = get_pwd("Enter database password: ", ctx);
    if (plen < 0) {
        warnusys("open ", ESC, db_file(ctx), ESC, ": cannot read password");
        return -1;
    }

    u32 magic = SSP_MAGIC;
    u64 ver;
    unsigned algo, iter;
    if (!shldata_read(&magic, &ver, &algo, &iter, &ctx->sa, AT_FDCWD, db_file(ctx), ctx->pwd, plen)) {
        if (errno == EINVAL && magic != SSP_MAGIC)
            warnu("open database ", ESC, db_file(ctx), ESC, ": ", "not an SSP database");
        else if (errno == EBADMSG)
            warnu("open database ", ESC, db_file(ctx), ESC, ": ", "wrong password");
        else
            warnusys("open ", ESC, db_file(ctx), ESC);
        return -1;
    }

    if (ver > 0) {
        warnu("open database ", ESC, db_file(ctx), ESC, ": ", "version too recent");
        return -1;
    }

    if (!ctx->iter) {
        ctx->iter = iter;
        dbg("set iteration to ", PMUINT(ctx->iter));
    }

    ctx->cdboff = off;
    cdb_init_frommem(&ctx->cdb, ctx->sa.s + off, ctx->sa.len - off);
    dbg("initialized cdb [off=", PMUINT(off), " size=", PMUINT(ctx->cdb.size), "]");

#if ! LIMB_IS_LE
    db_reset(ctx);
    while (db_next(ctx) == 1) {
        u64p_le(&((struct otp *) db_otp(ctx))->slen);
        u64p_le(&((struct otp *) db_otp(ctx))->n);
    }
#endif

    return 1;
}

void
close_db(struct ssp *ctx)
{
    dbg("closing database");

    if (ctx->cdb.map) {
        stralloc_remove(&ctx->sa, ctx->cdboff, ctx->cdb.size);
        ctx->cdboff = 0;
        ctx->cdb.map = NULL;
        ctx->cdb.size = 0;
    }
}

int
rebuild_cdb(cdbmaker_sa *mkr, const char *oldkey, const char *newkey,
            const struct otp *otp, struct ssp *ctx)
{
    size_t olen = 0, nlen = 0, vlen = 0;
    int r, n = 0;

    dbg("rebuiling cdb");

    /* oldkey & newkey  : renaming/editing an entry
     * oldkey & NULL    : removing an entry
     * NULL   & newkey  : adding an entry
     * NULL   & NULL    : isn't allowed
     */

    if (newkey) {
        nlen = strlen(newkey) + 1;
        vlen = sizeof(*otp) + otp->slen + strlen(otp->data + otp->slen) + 1;
    }
    if (oldkey) olen = strlen(oldkey) + 1;

    dbg("oldkey=", ESC, oldkey, ESC, " newkey=", ESC, newkey, ESC, " vlen=", PMUINT(vlen));

#if LIMB_IS_LE
    char *data = (char *) otp;
#else
    char data[vlen];
    memcpy(data, otp, vlen);
    u64p_le(&((struct otp *) data)->slen);
    u64p_le(&((struct otp *) data)->n);
#endif

    if (!cdbmaker_sa_start(mkr))
        return -1;
    /* new file? */
    if (!ctx->cdb.map) {
        dbg("new file, adding newkey");
        if (nlen && !cdbmaker_sa_add(mkr, newkey, nlen, data, vlen))
            return -1;
        ++n;
    } else {
        int a = !nlen;
        db_reset(ctx);
        dbg("iterating current cdb");
        while ((r = db_next(ctx)) == 1) {
            dbg("key=", ESC, db_entry(ctx), ESC);
            /* existing newkey isn't allowed when adding or renaming */
            if (nlen == ctx->key.len && !strcmp(newkey, ctx->key.s)
                    && (!olen || strcmp(oldkey, newkey))) {
                warnu("add entry ", ESC, newkey, ESC, ": Entry already exists");
                return (errno = EINVAL, -1);
            } else {
                /* Note the use of strcoll() here to ensure we order entries
                 * using locale-specific rules. */
                if (!a && strcoll(ctx->key.s, newkey) > 0) {
                    dbg("inserting newkey before");
                    /* insert otp at the right place */
                    if (!cdbmaker_sa_add(mkr, newkey, nlen, data, vlen))
                        return -1;
                    ++n;
                    a = 1;
                }
                /* if current entry isn't the old one nor the new one, re-add */
                if (!(olen == ctx->key.len && !strcmp(oldkey, ctx->key.s))
                        && !(nlen == ctx->key.len && !strcmp(newkey, ctx->key.s))) {
                    dbg("preserving key");
                    if (!cdbmaker_sa_add(mkr, ctx->key.s, ctx->key.len,
                                ctx->val.s, ctx->val.len))
                        return -1;
                    ++n;
                }
            }
        }
        if (r < 0) {
            warnu("read database: file corrupted");
            return (errno = EINVAL, -1);
        }
        if (!a) {
            dbg("inserting newkey last");
            /* insert last */
            if (!cdbmaker_sa_add(mkr, newkey, nlen, data, vlen))
                return -1;
            ++n;
        }
    }
    dbg("finishing rebuild (", PMINT(n), " entries)");
    if (!cdbmaker_sa_finish(mkr))
        return -1;

    return n;
}

int
write_db(char *data, size_t dlen, struct ssp *ctx)
{
    dbg("writing db; iter=", PMUINT((ctx->iter) ? ctx->iter : ITER_DEF));

    ssize_t plen;

    plen = get_pwd("Enter new database password: ", ctx);
    if (plen < 0) {
        warnusys("read password");
        return (errno = EINVAL, 0);
    }

    struct iovec v;
    v.iov_base = data;
    v.iov_len = dlen;

    if (!shldata_write(AT_FDCWD, db_file(ctx), SSP_MAGIC, 0, ctx->pwd, plen,
                ALGO_SHA3_256, (ctx->iter) ? ctx->iter : ITER_DEF, 1, &v, 1)) {
        warnusys("write database to ", ESC, db_file(ctx), ESC);
        return 0;
    }

    dbg("database written to ", ESC, db_file(ctx), ESC);
    return 1;
}

void
rebuild_db(const char *oldkey, const char *newkey, const struct otp *otp, struct ssp *ctx)
{
    dbg("rebuilding database");

    cdbmaker_sa mkr = CDBMAKER_SA_ZERO;

    int r = rebuild_cdb(&mkr, oldkey, newkey, otp, ctx);
    if (r < 0)
        diefu(exitcode_from_errno(errno), "save database",
              (errno == ENOMEM) ? ": Out of memory" : NULL);

    if (!r) {
        dbg("empty cdb, removing database");
        if (unlink(db_file(ctx)) < 0 && errno != ENOENT)
            diefusys(EX_IOERR, "remove database");
        out("Database empty, ", ESC, db_file(ctx), ESC, " removed");
    } else {
        dbg("saving new database");
        if (!write_db(mkr.sa.s, mkr.sa.len, ctx))
            diefusys((errno == EINVAL) ? EX_DATA_ERR : EX_CANTCREAT, "save database");
        out("Database saved to ", ESC, db_file(ctx), ESC);
    }

    cdbmaker_sa_free(&mkr);
}
