/* This file is part of ssp                             https://lila.oss/ssp
 * Copyright (C) 2023 Olivier Brunel                          jjk@jjacky.com */
/* SPDX-License-Identifier: GPL-2.0-only */
#include <errno.h>
#include <time.h>
#include <limb/base32.h>
#include <limb/command.h>
#include <limb/exitcode.h>
#include <limb/hasher.h>
#include <limb/hasher_sha1.h>
#include <limb/hasher_sha256.h>
#include <limb/hasher_sha512.h>
#include <limb/hasher_sha3_224.h>
#include <limb/hasher_sha3_256.h>
#include <limb/hasher_sha3_512.h>
#include <limb/hasher_blake3.h>
#include <limb/loadopt.h>
#include <limb/output.h>
#include <limb/stralloc.h>
#include <limb/u32.h>
#include <limb/u64.h>
#include "ssp.h"
#include "hotp.h"

struct get {
    struct otp otp;
    const char *secret;
    struct ssp *ssp;
    u64 ts;
    u8 upd_entry;
};

COMMAND(get, "Get a One-Time Password", "-e ENTRY | -s SECRET [OPTION..]",
" -a, --algo ALGO                       Set ALGO as hashing algorithm [sha1]\n"
" -C, --counter-val NUM                 Same as --counter but don't update entry\n"
" -c, --counter[=NUM]                   Set to counter-based (HOTP) with counter of NUM [1]\n"
" -d, --digits NUM                      Set to return an OTP of NUM digits [6]\n"
" -e, --entry ENTRY                     Get a password for ENTRY\n"
" -s, --secret SECRET                   Set SECRET as entry's secret\n"
" -T, --use-time TS                     Use TS as unix timestamp instead of current time\n"
" -t, --time[=SECS]                     Set to time-based (TOTP) with precision of SECS [30]\n"
);

static int
parse_cmdline(int argc, const char *argv[], const char usage[], struct get *ctx)
{
    const struct option options[] = {
        OPTION_ARG_REQ ('a', "algo",                    0,          OPTID_SHORTOPT),
        OPTION_ARG_REQ ('C', "counter",                 0,          OPTID_SHORTOPT),
        OPTION_ARG_OPT ('c', "counter",                 0,          OPTID_SHORTOPT),
        OPTION_ARG_REQ ('d', "digits",                  0,          OPTID_SHORTOPT),
        OPTION_ARG_REQ ('e', "entry",                   OPT_PATH,   OPTID_SHORTOPT),
        OPTION_ARG_REQ ('s', "secret",                  0,          OPTID_SHORTOPT),
        OPTION_ARG_REQ ('T', "use-time",                0,          OPTID_SHORTOPT),
        OPTION_ARG_OPT ('t', "time",                    0,          OPTID_SHORTOPT),
        LOADOPT_DONE
    };
    struct loadopt lo = LOADOPT_ZERO;

    int c, r;
    while ((c = loadopt(&ctx->ssp->sa, argc, argv, options, 0, NULL, 0, &lo))) switch (c) {
        case 'a':
            {
                int first = -1;
                r = validate_algo(&first, LO_ARG(&lo), strlen(LO_ARG(&lo)));
                if (r < 0) {
                    if (first > -1)
                        list_matches(obuffer_1, OLVL_NORMAL, "did you mean ",
                                     NULL, " or ", " ?", LO_ARG(&lo), strlen(LO_ARG(&lo)),
                                     first, (const char **) algos);
                    diecmdusage(EX_USAGE, usage, &command_get);
                }
                ctx->otp.algo = r;
                dbg("setting algo to ", algos[r]);
            }
            break;
        case 'C':
            ctx->upd_entry = 0;
            dbg("will not update entry's counter value");
            /* fall through */
        case 'c':
            if (!LO_ARG(&lo)) {
                ctx->otp.n = 1;
            } else {
                if (validate_counter(&ctx->otp.n, LO_ARG(&lo), strlen(LO_ARG(&lo))) < 0)
                    diecmdusage(EX_USAGE, usage, &command_get);
            }
            ctx->otp.type = TYPE_COUNTER;
            dbg("setting to HOTP with counter=", PMUINT(ctx->otp.n));
            break;
        case 'd':
            r = validate_digits(LO_ARG(&lo), strlen(LO_ARG(&lo)));
            if (r < 0)
                diecmdusage(EX_USAGE, usage, &command_get);
            ctx->otp.digits = r;
            dbg("setting to ", PMUINT(r), " digits");
            break;
        case 'e':
            {
                const char *entry = ctx->ssp->sa.s + LO_OFF(&lo);
                int r;

                dbg("looking for entry ", ESC, entry, ESC);

                if (ctx->ssp->cdb.map) {
                    /* re-align cdb in case of a re-alloc */
                    ctx->ssp->cdb.map = ctx->ssp->sa.s + ctx->ssp->cdboff;
                } else {
                    r = open_db(ctx->ssp, 1);
                    if (r < 0)
                        diefu(exitcode_from_errno(errno), "open database");
                    /* re-align */
                    entry = ctx->ssp->sa.s + LO_OFF(&lo);
                }

                r = db_find(entry, ctx->ssp);
                if (r < 0)
                    diefu(EX_DATA_ERR, "find entry ", ESC, entry, ESC);
                if (!r)
                    dief(EX_DATA_ERR, "no entry ", ESC, entry, ESC, " in database");

                ctx->otp = *db_otp(ctx->ssp);
                ctx->secret = db_otp(ctx->ssp)->data;
                ctx->upd_entry = 1;
                dbg("using entry ", ESC, entry, ESC);
            }
            break;
        case 's':
            ctx->secret = LO_ARG(&lo);
            ctx->otp.slen = 0;
            dbg("setting secret to ", ctx->secret);
            break;
        case 'T':
            if (!u64_scan0(&ctx->ts, LO_ARG(&lo))) {
                warn("invalid time stamp: ", LO_ARG(&lo));
                diecmdusage(EX_USAGE, usage, &command_get);
            }
            dbg("using time value of ", PMUINT(ctx->ts));
            break;
        case 't':
            if (!LO_ARG(&lo)) {
                ctx->otp.n = 30;
            } else {
                r = validate_precision(LO_ARG(&lo), strlen(LO_ARG(&lo)));
                if (r < 0)
                    diecmdusage(EX_USAGE, usage, &command_get);
                ctx->otp.n = r;
            }
            ctx->otp.type = TYPE_TIME;
            dbg("setting to TOTP with precision=", PMUINT(ctx->otp.n));
            break;

        case -1:
            diecmdusage(EX_USAGE, usage, &command_get);
        default:
            die(EX_SOFTWARE, "unexpected return value ", PMINT(c), " from loadopt");
    };

    return LO_CUR(&lo);
}

int
get_main(int argc, const char *argv[], const char *env[], const char usage[], void *ssp_)
{
    struct ssp *ssp = ssp_;
    struct get ctx = { .otp = OTP_DEFAULT, .secret = NULL, .ssp = ssp,
        .ts = (u64) -1, .upd_entry = 0 };

    (void) env;
    parse_cmdline(argc, argv, usage, &ctx);

    if (!ctx.secret)
        dief(EX_DATA_ERR, "secret missing");

    if (!ctx.otp.slen) {
        dbg("decoding secret from base32");
        size_t l = strlen(ctx.secret);
        ssize_t r = base32_scan(NULL, ctx.secret, l, 0);
        if (r < 0)
            dief(EX_DATA_ERR, "invalid secret: Not in base32");
        if (!stralloc_readyplus(&ssp->sa, r))
            diefusys(EX_TEMPFAIL, "process secret");
        base32_scan(ssp->sa.s + ssp->sa.len, ctx.secret, l , 0);
        ctx.secret = ssp->sa.s + ssp->sa.len;
        /* we don't update ssp->sa.len on purpose, so in case we need to rewrite
         * the db the calculation is easier */
        ctx.otp.slen = r;
    }

    hasher *hr = sha1;
    switch (ctx.otp.algo) {
        case ALGO_SHA256:   hr = sha256;   break;
        case ALGO_SHA512:   hr = sha512;   break;
        case ALGO_SHA3_224: hr = sha3_224; break;
        case ALGO_SHA3_256: hr = sha3_256; break;
        case ALGO_SHA3_512: hr = sha3_512; break;
        case ALGO_BLAKE3:   hr = blake3;   break;
    }

    u64 n;
    if (ctx.otp.type == TYPE_TIME) {
        if (ctx.ts == (u64) -1)
            ctx.ts = time(NULL);
        n = ctx.ts / ctx.otp.n;
    } else {
        n = ctx.otp.n;
    }

    dbg("calculating HOTP n=", PMUINT(n), " digits=", PMUINT(ctx.otp.digits));
    int r = hotp(hr, ctx.secret, ctx.otp.slen, n, ctx.otp.digits);
    add((ctx.otp.type == TYPE_COUNTER) ? "H" : "T", "OTP: ");
    char buf[ctx.otp.digits + 1];
    buf[u32_0fmt(buf, r, sizeof(buf) - 1)] = 0;
    quiet(buf);

    /* TYPE_COUNTER implies to write the db with an updated/incremented
     * counter (for next time), unless asked not to */
    if (ctx.upd_entry && ctx.otp.type == TYPE_COUNTER
            /* make sure the entry is also in COUNTER mode */
            && db_otp(ssp)->type == TYPE_COUNTER) {
        struct otp *e = (struct otp *) db_otp(ssp);
        e->n = ++ctx.otp.n;
        dbg("updating entry's counter to ", PMUINT(e->n));

        if (!write_db(ssp->sa.s + ssp->cdboff, ssp->cdb.size, ssp))
            diefusys((errno == EINVAL) ? EX_DATA_ERR : EX_CANTCREAT, "save database");
    }

    return 0;
}
