/* This file is part of ssp                             https://lila.oss/ssp
 * Copyright (C) 2023 Olivier Brunel                          jjk@jjacky.com */
/* SPDX-License-Identifier: GPL-2.0-only */
#include <errno.h>
#include <limb/base32.h>
#include <limb/bytestr.h>
#include <limb/cdbmake.h>
#include <limb/command.h>
#include <limb/copa.h>
#include <limb/djbunix.h>
#include <limb/esc.h>
#include <limb/exitcode.h>
#include <limb/hasher.h> /* ALGO_* */
#include <limb/loadopt.h>
#include <limb/output.h>
#include <limb/stralloc.h>
#include <limb/u16.h>
#include "ssp.h"

struct import {
    struct ssp *ssp;
    size_t fileoff;
};

enum {
    ARGID_FILE = OPTID_FIRST,
};

COMMAND(import, "Import entries from file", "<file>", NULL);

static int
parse_cmdline(int argc, const char *argv[], const char usage[], struct import *ctx)
{
    const struct option options[] = {
        LOADOPT_ARGUMENTS,
        ARGUMENT_REQ( "file",                           OPT_PATH,   ARGID_FILE),
        LOADOPT_DONE
    };
    struct loadopt lo = LOADOPT_ZERO;

    int c;
    while ((c = loadopt(&ctx->ssp->sa, argc, argv, options, 0, NULL, 0, &lo))) switch (c) {

        case ARGID_FILE:
            ctx->fileoff = LO_OFF(&lo);
            break;

        case -1:
            diecmdusage(EX_USAGE, usage, &command_import);
        default:
            die(EX_SOFTWARE, "unexpected return value ", PMINT(c), " from loadopt");
    };

    return LO_CUR(&lo);
}

static ssize_t
unesc(char *dst, size_t dlen, const char *sce, size_t slen)
{
    /* unescape it */
    ssize_t l = esc_scan(dst, dlen, sce, slen);
    if (l < 0) return -1;
    /* make sure there was no encoded NUL byte */
    if (byte_chr(dst, l, 0) < (size_t) l)
        return (errno = EINVAL, -1);
    dst[l++] = 0;
    return l;
}

int
import_main(int argc, const char *argv[], const char *env[], const char usage[], void *ssp_)
{
    struct ssp *ssp = ssp_;
    struct import ctx = { .ssp = ssp_, .fileoff = 0 };
    stralloc sacopa = STRALLOC_ZERO;
    stralloc saotp = STRALLOC_ZERO;

    (void) env;
    parse_cmdline(argc, argv, usage, &ctx);

    int fd = open_parsed_name(ssp->sa.s + ctx.fileoff, openb_read);
    if (fd < 0)
        diefusys(EX_NOINPUT, "open ", ESC, ssp->sa.s + ctx.fileoff, ESC);

    dbg("reading ", ESC, ssp->sa.s + ctx.fileoff, ESC);
    if (!slurp(&sacopa, fd))
        diefusys(exitcode_from_errno(errno), "read ", ESC, ssp->sa.s + ctx.fileoff, ESC);

    cdbmaker_sa mkr = CDBMAKER_SA_ZERO;
    struct otp otp_default = OTP_DEFAULT;
    struct otp otp;
    const char *name = NULL;
    size_t nlen = 0;
    const char *secret;
    ssize_t slen;
    size_t slen32;
    const char *comments;
    size_t clen;
    struct copa copa;
    int r, n = 0;
    dbg("init parsing");
    otp = otp_default;
    comments = NULL; clen = 0; /* silence warnings */
    copa_init(sacopa.s, sacopa.len, &copa);
    while ((r = copa_next(&copa)) >= 0) {
        dbg("r=", PMINT(r));
        if (r) dbg("is_section=", PMINT(copa_is_section(&copa)),
                   " name=", PMLEN(copa_name(&copa), copa_nlen(&copa)));
        if (!r || copa_is_section(&copa)) {
            if (name) {
                if (!slen)
                    die(EX_DATA_ERR, "missing secret in section/entry ",
                        ESC, PMLEN(name, nlen), ESC);

                dbg("preparing entry ", ESC, PMLEN(name, nlen), ESC);
                size_t otplen = sizeof(otp) + slen + clen + 1;
                if (!stralloc_ready(&saotp, otplen))
                    diefusys(EX_TEMPFAIL, "prepare entry ", ESC, PMLEN(name, nlen), ESC);

                struct otp *p = (struct otp *) saotp.s;
                memcpy(p, &otp, sizeof(otp));
                p->slen = slen;
                base32_scan(p->data, secret, slen32, 0);
                if (clen >= 2 && comments[0] == '"' && comments[clen - 1] == '"') {
                    ++comments;
                    clen -= 2;
                    dbg("unescaping comments");
                    if (unesc(p->data + slen, clen, comments, clen) < 0)
                        diefusys(EX_DATA_ERR, "read comments in section/entry ",
                                 ESC, PMLEN(name, nlen), ESC);
                } else {
                    memcpy(p->data + slen, comments, clen + 1);
                }

                /* if we've already imported an entry, switch cdb :
                 * - our newly generated one becomes the active db,
                 * - and we'll reuse the memory for our maker
                 */
                if (mkr.sa.s) {
                    dbg("making previously generated cdb live");
                    stralloc sa = ssp->sa;
                    ssp->sa = mkr.sa;
                    /* insert back db name (& import file name) */
                    if (!stralloc_insertb(&ssp->sa, 0, sa.s, ssp->cdboff))
                        diefusys(EX_TEMPFAIL, "prepare for importation");
                    cdb_init_frommem(&ssp->cdb, ssp->sa.s + ssp->cdboff,
                                     ssp->sa.len);
                    mkr.sa = sa;
                }

                /* open db if we haven't already */
                if (!ssp->cdb.map && open_db(ssp, 0) < 0)
                    diefu(exitcode_from_errno(errno), "open database");

                char key[nlen + 1];
                if (unesc(key, nlen, name, nlen) < 0)
                    dief(EX_DATA_ERR, "invalid section/entry name: ", PMLEN(name, nlen));
                out("Importing ", ESC, key, ESC, "...");
                if (rebuild_cdb(&mkr, NULL , key, p, ssp) < 0)
                    diefu(exitcode_from_errno(errno), "import data",
                          (errno == ENOMEM) ? ": Out of memory" : NULL);
                ++n;
            }
            if (!r) break;

            dbg("resetting new entry");
            name = copa_name(&copa);
            nlen = copa_nlen(&copa);
            otp = otp_default;
            comments = "";
            slen = clen = 0;
        } else if (!name) {
            dief(EX_DATA_ERR, "data before section/entry");
        } else {
            int r;

            /* all options must have a value */
            if (!copa_vlen(&copa))
                dief(EX_DATA_ERR, "no value for option ",
                     ESC, PMLEN(copa_name(&copa), copa_nlen(&copa)), ESC,
                     " in section/entry ", ESC, PMLEN(name, nlen), ESC);

            dbg("value=", ESC, PMLEN(copa_value(&copa), copa_vlen(&copa)), ESC);

            if (copa_nlen(&copa) == 4 && !strncmp(copa_name(&copa), "algo", 4)) {
                r = validate_algo(NULL, copa_value(&copa), copa_vlen(&copa));
                if (r < 0) goto err;
                otp.algo = r;
            } else if (copa_nlen(&copa) == 6 && !strncmp(copa_name(&copa), "digits", 6)) {
                r = validate_digits(copa_value(&copa), copa_vlen(&copa));
                if (r < 0) goto err;
                otp.digits = r;
            } else if (copa_nlen(&copa) == 7 && !strncmp(copa_name(&copa), "counter", 7)) {
                if (validate_counter(&otp.n, copa_value(&copa), copa_vlen(&copa)) < 0)
                    goto err;
            } else if (copa_nlen(&copa) == 4 && !strncmp(copa_name(&copa), "time", 4)) {
                if (copa_vlen(&copa) == 1 && copa_value(&copa)[0] == '1')
                    otp.type = TYPE_TIME;
                else if (copa_vlen(&copa) == 1 && copa_value(&copa)[0] == '0')
                    otp.type = TYPE_COUNTER;
                else
                    dief(EX_DATA_ERR, "invalid value for \"time\" in section/entry ",
                         ESC, PMLEN(name, nlen), ESC, ": ",
                         PMLEN(copa_value(&copa), copa_vlen(&copa)));
            } else if (otp.type == TYPE_TIME && copa_nlen(&copa) == 9
                    && !strncmp(copa_name(&copa), "precision", 9)) {
                r = validate_precision(copa_value(&copa), copa_vlen(&copa));
                if (r < 0) goto err;
                otp.n = r;
            } else if (copa_nlen(&copa) == 6 && !strncmp(copa_name(&copa), "secret", 6)) {
                slen = base32_scan(NULL, copa_value(&copa), copa_vlen(&copa), 0);
                if (slen < 0)
                    dief(EX_DATA_ERR, "invalid secret in section/entry ",
                         ESC, PMLEN(name, nlen), ESC, ": Not in base32: ",
                         PMLEN(copa_value(&copa), copa_vlen(&copa)));
                secret = copa_value(&copa);
                slen32 = copa_vlen(&copa);
            } else if (copa_nlen(&copa) == 8 && !strncmp(copa_name(&copa), "comments", 8)) {
                comments = copa_value(&copa);
                clen = copa_vlen(&copa);
            } else {
                dief(EX_DATA_ERR, "invalid data, unknown value ",
                     ESC, PMLEN(copa_name(&copa), copa_nlen(&copa)), ESC,
                     " in section/entry ", ESC, PMLEN(name, nlen), ESC);
            }
        }
    }
    if (r < 0)
        diefu(EX_DATA_ERR, "parse ", ESC, ssp->sa.s + ctx.fileoff, ESC);

    if (mkr.sa.s && !write_db(mkr.sa.s, mkr.sa.len, ssp))
        diefusys((errno == EINVAL) ? EX_DATA_ERR : EX_CANTCREAT, "save database");
    cdbmaker_sa_free(&mkr);

    if (!n)
        dief(EX_DATA_ERR, "nothing to import");
    out("Successfully imported ", PMUINT(n), " entries to ", ESC, db_file(ssp), ESC);

    stralloc_free(&sacopa);
    stralloc_free(&saotp);
    return 0;

err:
    dief(EX_DATA_ERR, "invalid section/entry ", ESC, PMLEN(name, nlen), ESC);
}
