/* This file is part of ssp                             https://lila.oss/ssp
 * Copyright (C) 2023 Olivier Brunel                          jjk@jjacky.com */
/* SPDX-License-Identifier: GPL-2.0-only */
#include <errno.h>
#include <fnmatch.h>
#include <limb/command.h>
#include <limb/exitcode.h>
#include <limb/loadopt.h>
#include <limb/output.h>
#include "ssp.h"

enum {
    OPT_DETAILS         = 1 << 0,
    OPT_SECRET          = 1 << 1, /* must be the same as in show */
    OPT_NO_COMMENTS     = 1 << 2, /* must be the same as in show */
    OPT_ESC_COMMENTS    = 1 << 3, /* must be the same as in show */
    OPT_FMT_INI         = 1 << 4,
};

enum {
    ARGID_PATTERN = OPTID_FIRST,
};

struct list {
    struct ssp *ssp;
    const char *sep;
    size_t ptrnoff;
    unsigned options;
};

COMMAND(list, "List all entries", "[OPTION..] [<pattern>]",
" -C, --no-comments                     Do not show comments with --details/--format\n"
" -d, --details                         Show entries' details\n"
" -e, --escape-comments                 Escape comments on a single line (implies --details)\n"
" -f, --format                          Output in INI-like format\n"
" -S, --secret                          Show entries' secrets (implies --details)\n"
" -s, --sep SEP                         Use SEP as separator\n"
);

static int
parse_cmdline(int argc, const char *argv[], const char usage[], struct list *ctx)
{
    const struct option options[] = {
        OPTION_ARG_NONE('C', "no-comments",             0,          OPTID_SHORTOPT),
        OPTION_ARG_NONE('d', "details",                 0,          OPTID_SHORTOPT),
        OPTION_ARG_NONE('e', "escape-comments",         0,          OPTID_SHORTOPT),
        OPTION_ARG_NONE('f', "format",                  0,          OPTID_SHORTOPT),
        OPTION_ARG_NONE('S', "secret",                  0,          OPTID_SHORTOPT),
        OPTION_ARG_REQ ('s', "sep",                     0,          OPTID_SHORTOPT),
        LOADOPT_ARGUMENTS,
        ARGUMENT_OPT( "pattern",                        OPT_PATH,   ARGID_PATTERN),
        LOADOPT_DONE
    };
    struct loadopt lo = LOADOPT_ZERO;

    int c;
    while ((c = loadopt(&ctx->ssp->sa, argc, argv, options, 0, NULL, 0, &lo))) switch (c) {
        case 'C':
            ctx->options |= OPT_NO_COMMENTS;
            break;
        case 'd':
            ctx->options |= OPT_DETAILS;
            break;
        case 'e':
            ctx->options |= OPT_DETAILS | OPT_ESC_COMMENTS;
            break;
        case 'f':
            ctx->options |= OPT_FMT_INI;
            break;
        case 'S':
            ctx->options |= OPT_DETAILS | OPT_SECRET;
            break;
        case 's':
            ctx->sep = LO_ARG(&lo);
            break;

        case ARGID_PATTERN:
            ctx->ptrnoff = LO_OFF(&lo);
            break;

        case -1:
            diecmdusage(EX_USAGE, usage, &command_list);
        default:
            die(EX_SOFTWARE, "unexpected return value ", PMINT(c), " from loadopt");
    };

    return LO_CUR(&lo);
}

int
list_main(int argc, const char *argv[], const char *env[], const char usage[], void *ssp_)
{
    struct ssp *ssp = ssp_;
    struct list ctx = { .ssp = ssp, .sep = "\n", .ptrnoff = (size_t) -1 };

    (void) env;
    parse_cmdline(argc, argv, usage, &ctx);

    int r = open_db(ssp, 1);
    if (r < 0)
        diefu(exitcode_from_errno(errno), "open database");

    const char *ptrn = NULL;
    if (ctx.ptrnoff != (size_t) -1)
        ptrn = ssp->sa.s + ctx.ptrnoff;

    int i = 0;
    db_reset(ssp);
    dbg("iterating database...");
    while ((r = db_next(ssp)) == 1) {
        dbg("entry ", ESC, db_entry(ssp), ESC);
        if (ptrn && fnmatch(ptrn, db_entry(ssp), 0))
            continue;
        if (ctx.options & OPT_FMT_INI) {
            show_entry_ini(ctx.options & OPT_NO_COMMENTS, ssp);
        } else if (ctx.options & OPT_DETAILS) {
            show_entry(ctx.options & (OPT_NO_COMMENTS | OPT_ESC_COMMENTS | OPT_SECRET), ssp);
        } else {
            if (i) add(ctx.sep);
            add(ESC, db_entry(ssp), ESC);
        }
        i = 1;
    }
    if (r < 0)
        diefu(EX_DATA_ERR, "read database: file corrupted");
    if (!i)
        dief(EX_DATA_ERR, "no entry matching ", ESC, ptrn, ESC);

    /* add new line + flush buffers */
    if (!(ctx.options & (OPT_DETAILS | OPT_FMT_INI))) out(NULL);

    return 0;
}
