/* This file is part of ssp                             https://lila.oss/ssp
 * Copyright (C) 2023 Olivier Brunel                          jjk@jjacky.com */
/* SPDX-License-Identifier: GPL-2.0-only */
#include <errno.h>
#include <limb/base32.h>
#include <limb/bytestr.h>
#include <limb/command.h>
#include <limb/exitcode.h>
#include <limb/hasher.h> /* ALGO_* */
#include <limb/loadopt.h>
#include <limb/output.h>
#include "ssp.h"

enum {
    OPT_SECRET          = 1 << 1, /* must be the same in list */
    OPT_NO_COMMENTS     = 1 << 2, /* must be the same in list */
    OPT_ESC_COMMENTS    = 1 << 3, /* must be the same in list */
    OPT_FMT_INI         = 1 << 4,
};

struct show {
    struct ssp *ssp;
    size_t entoff;
    unsigned options;
};

enum {
    ARGID_ENTRY = OPTID_FIRST,
};

COMMAND(show, "Show an entry", "[OPTION..] <entry>",
" -C, --no-comments                     Do not show entry's comments\n"
" -e, --escape-comments                 Escape comments on a single line\n"
" -f, --format                          Output in INI-like format. Implies --secret\n"
" -S, --secret                          Show entry's secret\n"
);

static int
parse_cmdline(int argc, const char *argv[], const char usage[], struct show *ctx)
{
    const struct option options[] = {
        OPTION_ARG_NONE('C', "no-comments",             0,          OPTID_SHORTOPT),
        OPTION_ARG_NONE('e', "escape-comments",         0,          OPTID_SHORTOPT),
        OPTION_ARG_NONE('f', "format",                  0,          OPTID_SHORTOPT),
        OPTION_ARG_NONE('S', "secret",                  0,          OPTID_SHORTOPT),
        LOADOPT_ARGUMENTS,
        ARGUMENT_REQ( "entry",                          OPT_PATH,   ARGID_ENTRY),
        LOADOPT_DONE
    };
    struct loadopt lo = LOADOPT_ZERO;

    int c;
    while ((c = loadopt(&ctx->ssp->sa, argc, argv, options, 0, NULL, 0, &lo))) switch (c) {
        case 'C':
            ctx->options |= OPT_NO_COMMENTS;
            break;
        case 'e':
            ctx->options |= OPT_ESC_COMMENTS;
            break;
        case 'f':
            ctx->options |= OPT_FMT_INI;
            break;
        case 'S':
            ctx->options |= OPT_SECRET;
            break;

        case ARGID_ENTRY:
            ctx->entoff = LO_OFF(&lo);
            break;

        case -1:
            diecmdusage(EX_USAGE, usage, &command_show);
        default:
            die(EX_SOFTWARE, "unexpected return value ", PMINT(c), " from loadopt");
    };

    return LO_CUR(&lo);
}

void
show_entry(unsigned options, struct ssp *ctx)
{
    const char *types[] = { "Counter", "Time" };
    const struct otp *otp = db_otp(ctx);
    out("- Entry ", ESC, db_entry(ctx), ESC);
    out("   Algo: ", algos[otp->algo]);
    add("   ", types[otp->type], "-based; ");
    if (otp->type == TYPE_COUNTER)
        out("Counter: ", PMUINT(otp->n));
    else /* TYPE_TIME */
        out("Precision: ", PMUINT(otp->n), "s");
    out("   Digits: ", PMUINT(otp->digits));
    if (options & OPT_SECRET) {
        size_t l = base32_fmt(NULL, otp->data, otp->slen, 0);
        char buf[l+1];
        base32_fmt(buf, otp->data, otp->slen, 0);
        buf[l] = 0;
        out("   Secret: ", buf);
    }
    if (!(options & OPT_NO_COMMENTS) && otp->data[otp->slen]) {
        add("   Comments:");
        if (options & OPT_ESC_COMMENTS) {
            out(" ", ESC, otp->data + otp->slen, ESC);
        } else {
            const char *s = otp->data + otp->slen;
            size_t l = strlen(s);
            out((byte_chr(s, l, '\n') < l) ? "\n" : " ", s);
        }
    }
}

void
show_entry_ini(unsigned options, struct ssp *ctx)
{
    const struct otp *otp = db_otp(ctx);

    out("[", PUTMSG_TOGGLE_ESC, db_entry(ctx), PUTMSG_TOGGLE_ESC, "]");
    out("algo=", algos[otp->algo]);
    out("digits=", PMUINT(otp->digits));
    if (otp->type == TYPE_TIME) {
        out("time=1");
        out("precision=", PMUINT(otp->n));
    } else {
        out("counter=", PMUINT(otp->n));
    }
    size_t l = base32_fmt(NULL, otp->data, otp->slen, 1);
    char buf[l + 1];
    base32_fmt(buf, otp->data, otp->slen, 1);
    buf[l] = 0;
    out("secret=", buf);
    if (!(options & OPT_NO_COMMENTS) && otp->data[otp->slen])
        out("comments=", ESC, otp->data + otp->slen, ESC);
}

int
show_main(int argc, const char *argv[], const char *env[], const char usage[], void *ssp_)
{
    struct ssp *ssp = ssp_;
    struct show ctx = { .ssp = ssp };

    (void) env;
    parse_cmdline(argc, argv, usage, &ctx);

    int r = open_db(ssp, 1);
    if (r < 0)
        diefu(exitcode_from_errno(errno), "open database");

    const char *entry = ssp->sa.s + ctx.entoff;
    r = db_find(entry, ssp);
    if (r < 0)
        diefu(EX_DATA_ERR, "find entry ", ESC, entry, ESC);
    if (!r)
        dief(EX_DATA_ERR, "no entry ", ESC, entry, ESC, " in database");

    if (ctx.options & OPT_FMT_INI)
        show_entry_ini(ctx.options & OPT_NO_COMMENTS, ssp);
    else
        show_entry(ctx.options, ssp);

    return 0;
}
